""" abstract base for python-based study executors """
from abc import ABC, abstractmethod


class BaseRunner(ABC):
    """Generic base class for python study runners.

    To use a python runner, extend this class exactly once in a new module.
    See ./studies/vocab for an example usage.
    """

    @abstractmethod
    def run_executor(self, cursor: object, schema: str, verbose: bool):
        """Main entrypoint for python runners

        :param cursor: A PEP-249 compatible cursor
        :param schema: A schema name
        :param verbose: toggle for verbose output mode
        """
        raise NotImplementedError



""" Module for generating dynamic denormalized table from FHIR resources. 

If there is a nested resource we're extracting (i.e. a column with multiple
extensions, or a codingConcept field that could have more than one value),
we attempt to dynamically generate denormalized tables of the subset of
the data we're interested in, and we'll later left join these tables to
the parent when we're building the core resource tables.
"""
import csv

from cumulus_library.base_runner import BaseRunner
from cumulus_library.helper import get_progress_bar, query_console_output
from cumulus_library.template_sql.templates import (
    get_codeable_concept_denormalize_query,
    get_extension_denormalize_query,
    ExtensionConfig,
)


class PatientExtensionRunner(BaseRunner):
    def run_executor(self, cursor: object, schema: str, verbose: bool):
        """input point from make.execute_sql_template.

        :param cursor: A database cursor object
        :param schema: the schema/db name, matching the cursor
        :param verbose: if true, outputs raw query, else displays progress bar
        """
        extension_types = [
            {
                "name": "race",
                "fhirpath": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-race",
            },
            {
                "name": "ethnicity",
                "fhirpath": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity",
            },
        ]
        configs = []
        for extension in extension_types:
            configs.append(
                ExtensionConfig(
                    "patient",
                    "id",
                    f"core__patient_ext_{extension['name']}",
                    extension["name"],
                    extension["fhirpath"],
                    ["ombCategory", "detailed", "text"],
                )
            )

        with get_progress_bar(disable=verbose) as progress:
            task = progress.add_task(
                f"Preprocessing core datasets...",
                total=len(configs) + 1,
                visible=not verbose,
            )
            self.build_patient_extensions(
                self, cursor, schema, verbose, progress, task, configs
            )
            self.build_condition_codes(self, cursor, schema, verbose, progress, task)

    @staticmethod
    def build_patient_extensions(
        self, cursor, schema, verbose, progress, task, configs
    ):
        """Constructs patient extension queries and posts to athena."""
        for config in configs:
            extension_query = get_extension_denormalize_query(config)
            cursor.execute(extension_query)
            query_console_output(verbose, extension_query, progress, task)

    @staticmethod
    def build_condition_codes(self, cursor, schema, verbose, progress, task):
        """Constructs patient extension queries and posts to athena."""
        codeable_concept_query = get_codeable_concept_denormalize_query(
            "condition",
            "code",
            "core__condition_codable_concepts",
            [
                "http://snomed.info/sct",
                "http://hl7.org/fhir/sid/icd-10-cm",
                "http://hl7.org/fhir/sid/icd-9-cm",
            ],
        )
        cursor.execute(codeable_concept_query)
        query_console_output(verbose, codeable_concept_query, progress, task)


if __name__ == "__main__":
    extension_types = [
        {
            "name": "race",
            "fhirpath": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-race",
        },
        {
            "name": "ethnicity",
            "fhirpath": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity",
        },
    ]
    configs = []
    for extension in extension_types:
        configs.append(
            ExtensionConfig(
                "patient",
                "id",
                f"core__patient_ext_{extension['name']}",
                extension["name"],
                extension["fhirpath"],
                ["ombCategory", "detailed", "text"],
            )
        )

    for config in configs:
        print(get_extension_denormalize_query(config))



""" Module for directly loading ICD bsvs into athena tables """
import csv

from pathlib import Path

from cumulus_library.base_runner import BaseRunner
from cumulus_library.helper import query_console_output, get_progress_bar
from cumulus_library.template_sql.templates import (
    get_ctas_query,
    get_insert_into_query,
)


class VocabIcdRunner(BaseRunner):
    def run_executor(self, cursor: object, schema: str, verbose: bool):
        self.create_icd_legend(self, cursor, schema, verbose)

    @staticmethod
    def clean_row(row, filename):
        """Removes non-SQL safe charatcers from the input row."""
        for i in range(len(row)):
            cell = str(row[i]).replace("'", "").replace(";", ",")
            row[i] = cell
        return row

    def create_icd_legend(
        self, cursor: object, schema: str, verbose: bool, partition_size: int = 1200
    ):
        """input point from make.execute_sql_template.

        :param cursor: A database cursor object
        :param schema: the schema/db name, matching the cursor
        :param verbose: if true, outputs raw query, else displays progress bar
        :partition_size: number of lines to read. Athena queries have a char limit.
        """
        table_name = "vocab__icd"
        icd_files = ["ICD10CM_2023AA", "ICD10PCS_2023AA", "ICD9CM_2023AA"]
        path = Path(__file__).parent
        query_count = 1  # accounts for static CTAS query
        for filename in icd_files:
            query_count += (
                sum(1 for i in open(f"{path}/{filename}.bsv", "rb")) / partition_size
            )

        with get_progress_bar(disable=verbose) as progress:
            task = progress.add_task(
                f"Uploading {table_name} data...",
                total=query_count,
                visible=not verbose,
            )
            self.build_vocab_icd(
                self,
                cursor,
                schema,
                verbose,
                partition_size,
                table_name,
                path,
                icd_files,
                progress,
                task,
            )

    @staticmethod
    def build_vocab_icd(
        self,
        cursor,
        schema,
        verbose,
        partition_size,
        table_name,
        path,
        icd_files,
        progress,
        task,
    ):
        """Constructs queries and posts to athena."""
        headers = ["CUI", "TTY", "CODE", "SAB", "STR"]
        header_types = [f"{x} string" for x in headers]
        rows_processed = 0
        dataset = []
        created = False
        for filename in icd_files:
            with open(f"{path}/{filename}.bsv", "r") as file:
                # For the first row in the dataset, we want to coerce types from
                # varchar(len(item)) athena default to to an unrestricted varchar, so
                # we'll create a table with one row - this make the recast faster, and
                # lets us set the partition_size a little higher by limiting the
                # character bloat to keep queries under athena's limit of 262144.
                reader = csv.reader(file, delimiter="|")
                if not created:
                    row = self.clean_row(next(reader), filename)
                    ctas_query = get_ctas_query(
                        schema_name=schema,
                        table_name=table_name,
                        dataset=[row],
                        table_cols=headers,
                    )
                    cursor.execute(ctas_query)
                    query_console_output(verbose, ctas_query, progress, task)
                    created = True
                for row in reader:
                    row = self.clean_row(row, filename)
                    dataset.append(row)
                    rows_processed += 1
                    if rows_processed == partition_size:
                        insert_into_query = get_insert_into_query(
                            table_name=table_name, table_cols=headers, dataset=dataset
                        )
                        query_console_output(verbose, insert_into_query, progress, task)
                        cursor.execute(insert_into_query)
                        dataset = []
                        rows_processed = 0
                if rows_processed > 0:
                    insert_into_query = get_insert_into_query(
                        table_name=table_name, table_cols=headers, dataset=dataset
                    )
                    cursor.execute(insert_into_query)
                    query_console_output(verbose, insert_into_query, progress, task)
